#include <cstdlib>
#include <iostream>
#include <fstream>
#include <string>
#include <algorithm>
#include <iomanip>

#include "HMAC-SHA2.h"

using namespace std;

unsigned long SHA2_initial_values[8] = {
	0x6a09e667L,
	0xbb67ae85L,
	0x3c6ef372L,
	0xa54ff53aL,
	0x510e527fL,
	0x9b05688cL,
	0x1f83d9abL,
	0x5be0cd19L
};

int main(int argc, char *argv[])
{
	cout << "This program can be used to hash the initial inner" << endl << "and outer keys for a HMAC-SHA2 implementation" << endl;
	char input;
	bool error = true;
	unsigned char key[64];

	while(error) {
		do {
			cout << endl;
			cout << "To type key in ASCII, press a and enter" << endl;
			cout << "To type key in hex, type h and enter" << endl;
			cout << "To quit, type q and enter" << endl;

			cin >> input;

			cin.ignore(1000,'\n');

		} while(input != 'h' &&
				input != 'H' &&
				input != 'a' &&
				input != 'A' &&
				input != 'q' &&
				input != 'Q');

		error = false;

		string tempkey;
		switch(input) {
			case 'h':
			case 'H':
				{
					cout << "Type the key in hex. Max 32 bytes (64 characters) Use lowercase letters" << endl;
					getline(cin,tempkey);
					// 傫̏ꍇALhexł͂܂B
					if(tempkey.size() & 0x01) {
						cout << "Not valid hex input, not an even number of characters" << endl;
						system("pause");
						error = true;
					} else if(tempkey.size() > 64) {
						cout << "Too long key" << endl;
						system("pause");
						error = true;
					} else {
						string::iterator iter;
						int temp;
						int i = 0;
						for(iter = tempkey.begin(); iter != tempkey.end() && !error; ++iter) {
							if(*iter >= 0x30 && *iter <= 0x39) {
								temp = (*iter - 0x30)*16;
							} else if (*iter >= 97 && *iter <= 102) {
								temp = (*iter - 87)*16;
							} else {
								cout << "Not a valid hex input, only use 0-9 and a-f" << endl;
								system("pause");
								error = true;
							}
							if(!error) {
								++iter;
								if(*iter >= 0x30 && *iter <= 0x39) {
									temp += ((int)*iter - 0x30);
								} else if (*iter >= 97 && *iter <= 102) {
									temp += ((int)*iter - 87);
								} else {
									cout << "Not a valid hex input, only use 0-9 and a-f" << endl;
									system("pause");
									error = true;
								}
								key[i++] = temp;
							}
						}
						for(; i < 64; ++i) {
							key[i] = 0x00;
						}
					}
					break;
				}

			case 'a':
			case 'A':
				{
					cout << "Type the key in ascii. Max 32 characters" << endl;
					getline(cin,tempkey);
					if(tempkey.size() > 32) {
						cout << "Too long key" << endl;
						system("pause");
						error = true;
					} else {
						string::iterator iter;
						int i = 0;
						for(iter = tempkey.begin(); iter != tempkey.end(); ++iter) {
							key[i++] = *iter;
						}
						for(; i < 64; ++i) {
							key[i] = 0x00;
						}
					}
					break;
				}

			case 'q':
			case 'Q':
				{
					return 0;
				}
		}
	}

	ofstream fout("HMAC-SHA2_keyData.h");
	fout.setf(ios::hex, ios::basefield);

	fout << "#ifndef HMAC_SHA2_KEYDATA" << endl;
	fout << "#define HMAC_SHA2_KEYDATA" << endl;
	fout << endl;
	fout << "// Calculated from the following secret key:" << endl;
	fout << "// In ASCII:\t";
	for(int i = 0; i < strlen((const char*)key); ++i ) {
		fout << key[i];
	}
	fout << endl;
	fout << "// In hex:   \t";
	for(int i = 0; i < 32; ++i) {
		if(i != 0 && i%4 == 0) {
			fout << " ";
		}
		fout << hex << setw(2) << setfill('0') << (int)key[i];
	}
	fout << endl << endl;

	// Kiɏ邽߂Ɍ̴ި݂
	unsigned char temp;
	for(int i = 0; i < 8; ++i) {
		temp = key[4*i + 3];
		key[4*i + 3] = key[4*i];
		key[4*i] = temp;

		temp = key[4*i + 2];
		key[4*i + 2] = key[4*i + 1];
		key[4*i + 1] = temp;
	}

    //****************
    // SHA2

	unsigned long inner_digest[8];
	unsigned long outer_digest[8];
	unsigned long ul_modified_key[16];
	unsigned char *uc_modified_key = (unsigned char*)ul_modified_key;

	for(int i = 0; i < 8; ++i) {
		inner_digest[i] = SHA2_initial_values[i];
	}

	for(int i = 0; i < 64; ++i) {
		uc_modified_key[i] = key[i] ^ 0x36;
	}

	SHA_2(ul_modified_key, inner_digest);

	cout << "The inner prehashed key: " << endl;
	cout.setf(ios::hex, ios::basefield);
	for(int i = 0; i < 8; ++i) {
		cout << "0x" << setw(8) << setfill('0') << inner_digest[i] << endl;
	}

	for(int i = 0; i < 8; ++i) {
		outer_digest[i] = SHA2_initial_values[i];
	}

	for(int i = 0; i < 64; ++i) {
		uc_modified_key[i] = key[i] ^ 0x5c;
	}

	SHA_2(ul_modified_key, outer_digest);

	cout << "The outer prehashed key: " << endl;
	cout.setf(ios::hex, ios::basefield);
	for(int i = 0; i < 8; ++i) {
		cout << "0x" << setw(8) << setfill('0') << outer_digest[i] << endl;
	}
	cout << endl;

	fout << "//! Data table containing the resulting hash from performing the inital SHA-2 operation on the secret key XOR 0x36" << endl;
	fout << "__flash uint32_t ful_inSHA[] = {" << endl;
	for(int i = 0; i < 7; ++i) {
		fout << "\t0x" << setw(8) << setfill('0') << inner_digest[i] << "," << endl;
	}
	fout << "\t0x" << setw(8) << setfill('0') << inner_digest[7] << "};" << endl;
	fout << endl;
	fout << "//! Data table containing the resulting hash from performing the inital SHA-2 operation on the secret key XOR 0x5c" << endl;
	fout << "__flash uint32_t ful_outSHA[] = {" << endl;
	for(int i = 0; i < 7; ++i) {
		fout << "\t0x" << setw(8) << setfill('0') << outer_digest[i] << "," << endl;
	}
	fout << "\t0x" << setw(8) << setfill('0') << outer_digest[7] << "};" << endl;
	fout << endl;

	fout << "#endif" << endl;
	fout.close();

	cout << "Both have been saved to the file HMAC-SHA2_keyData.h." << endl;
	cout << "Copy it to the folder ../lib_mcu/authentication/HMAC-SHA2/ and recompile the code to use the new key." << endl;

	system("pause");
	return EXIT_SUCCESS;
}
